###############################################################################
#
# Xylem K ratio functions
#
###############################################################################
"""
    xylem_k_ratio(vc::WeibullSingle{FT}, p_25::FT, vis::FT) where {FT<:AbstractFloat}
    xylem_k_ratio(vc::  WeibullDual{FT}, p_25::FT, vis::FT) where {FT<:AbstractFloat}

Returns the relative hydraulic conductance, given
- `vc` Xylem vulnerability curve
- `p` Xylem pressure at 298.15 K in `[MPa]`
- `p_25` Equivalent xylem pressure at 298.15 K in `[MPa]`
- `vis` Relative viscosity. If missing, vis = 1.
"""
function xylem_k_ratio(
            vc::WeibullSingle{FT},
            p_25::FT,
            vis::FT = FT(1)
) where {FT<:AbstractFloat}
    @unpack b,c = vc;

    if p_25<0
        kr = max( FT(1e-4), exp( -1 * (-p_25 / vc.b) ^ vc.c ) / vis );
    else
        kr = 1 / vis;
    end

    return kr
end




function xylem_k_ratio(
            vc::WeibullDual{FT},
            p_25::FT,
            vis::FT = FT(1)
) where {FT<:AbstractFloat}
    @unpack b1,c1,f1,b2,c2,f2 = vc;

    if p_25<0
        k1 = exp( -1 * (-p_25 / b1) ^ c1 ) * f1;
        k2 = exp( -1 * (-p_25 / b2) ^ c2 ) * f2;
        kr = max( FT(1e-4), (k1+k2) / vis );
    else
        kr = 1 / vis;
    end

    return kr
end
